@@ -1,12 +1,12 @@
-#!/usr/bin/python2.7
+#!/usr/bin/python3
 # -*- coding: utf-8 -*-
 #
 # Univention Lib
 #  python module
 #
-# Copyright 2011-2017 Univention GmbH
+# Copyright 2011-2021 Univention GmbH
 #
-# http://www.univention.de/
+# https://www.univention.de/
 #
 # All rights reserved.
 #
@@ -29,13 +29,11 @@
 # You should have received a copy of the GNU Affero General Public
 # License with the Debian GNU/Linux or Univention distribution in file
 # /usr/share/common-licenses/AGPL-3; if not, see
-# <http://www.gnu.org/licenses/>.
+# <https://www.gnu.org/licenses/>.
 
-from univention.lib.error import *
-from univention.lib.shell import *
-from univention.lib.locking import *
-from univention.lib.i18n import *
-try:
-	from univention.lib.misc import *
-except ImportError:
-	pass
+__path__ = __import__('pkgutil').extend_path(__path__, __name__)  # type: ignore
+
+__all__ = ('get_lock', 'release_lock', 'Translation', 'NullTranslation', 'Locale', 'I18N_Error')
+
+from univention.lib.locking import get_lock, release_lock
+from univention.lib.i18n import Translation, NullTranslation, Locale, I18N_Error
